/**
 *	Written 2010 by Tyler Montbriand.  Freeware.
 *	http://burningsmell.org/sensor8/
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "tcgets.h"
#include "sensor8.h"
#include "tsm_getopt.h"
#include "mydebug.h"

struct
{
	const char *dev;	// What dev to open, i.e. /dev/ttyUSB0
	const char *mode;	// What serial mode, probably 115200N81
	int verbose;	// Prints verbose debugging output when enabled.
	int usage;	// Prints usage when enabled.
	int raw;	// Prints raw bytes when enabled.
	int lock;	// Uses POSIX advisor locking when enabled.

	/**
	 *	If a sensor is present, its first read will produce ERR.
	 *	Subsequent reads to it, for a short time, might produce
	 *	a spurious temperature like 85 degrees C.
	 *	If the first read fails, it will try again at most
	 *	'trymax' number of times, waiting 'tryms' milliseconds
	 *	between each attempted read.
	 *
	 *	Anything below 500 milliseconds doesn't seem reliable.
	 *	Once they start producing sane numbers they seem to
	 *	keep producing sane numbers however.
	 */
	int tryms;
	/**
	 *	How many times to try to read a valid temp.
	 */
	int trymax;	// Max # of temp read retries
	int rdmin;	// Min # of valid temp reads
} globals={
	.dev	=	"/dev/ttyUSB0",
	.mode	=	"115200N81",
	.verbose=	0,
	.usage	=	0,
	.raw	=	0,
	.tryms	=	650,
	.trymax	=	1,
	.rdmin	=	3,
	.lock	=	0,
	};

const struct opt mopts[];

int open_dev(void), close_dev(int fd);

static int compare_s(const temp_t *a, const temp_t *b);

typedef int (*compare_func)(const void *a, const void *b);

/**
 *	Program to send commands to a DLP-I08 USB I/O module.
 */
int main(int argc, char *argv[])
{
	temp_t t;
	int fd=-1, arg;

	if(getenv("SENSOR8"))
		globals.dev=getenv("SENSOR8");

	arg=tsm_getopt(argc, argv, mopts);
	if((arg < 0) || globals.usage || (globals.rdmin < 1) ||
		(globals.rdmin >= 8))
	{
		fprintf(stderr, 
			"sensor8, utility to talk to a DLP-I08 module\n"
			"Written 2010 Tyler Montbriand.  Freeware.\n\n");
		tsm_usage(argc, argv, mopts);
		fprintf(stderr, "\nValid commands:\n"
			"\t[0-7]\t\tex. 7, reads logic value on pin 7\n"
			"\tV[0-7]\t\tex. V3, reads voltage on pin 3\n"
			"\tT[0-7]\t\tex. T5, reads Celsius on pin 5\n"
			"\tF[0-7]\t\tex. F3, reads Fahrenheit on pin 3\n"
			"\t[0-7]:[01]\tex. 1:0, sets logic 0 on pin 1\n"
			"\tZ[0-7]\t\tex. Z2, tristates pin 2\n"
			"\tACK\t\tPings device, waits for reply\n"
			"\tD[0-9]+\t\tex. D100, delays 100 milliseconds\n"
			"\tU[0-9]+\t\tex. U1000, delays 1000 microseconds\n\n"
			"\t%s ACK 1:1 D100 T0 Z1\n"
			"\t\t...will try to ACK the device, set pin 1 to logic high,\n"
			"\t\tsleep 100 milliseconds, read temp from pin 0, then\n"
			"\t\ttristate pin 1.\n\n"
			"\t\tPrepend @ to any command to make it silent.\n",
			argv[0] );
		return(1);
	}

	if(arg == argc)
	{
		fprintf(stderr, "No commands given\n");
		return(2);
	}

	if(globals.verbose)	setenv("DEBUG", "1", 1);


	fd=open_dev();

	if(fd < 0)
	{
		perror(globals.dev);
		return(1);
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_WRLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLKW, &l) != 0)
		{
			perror("Couldn't lock");
			close(fd);
			return(1);
		}
	}
#endif

	{
		// Set binary mode so we don't have to parse things like
		// "999.99*C\r\r"
		char c=CMD_SET_BINARY;
		write(fd, &c, 1);
	}


	arg--;
	while((++arg) < argc)
	{
		int pin=-1;
		unsigned short dat;
		unsigned char val, toss, cmd, silent=0, tries=0;
		const char *a=argv[arg];
		const char *pincmd=NULL;
		temp_t temp[8];
		int temp_pos=0;

		enum { TEMP_C=1, VOLT_IN=2, PIN_IN=3, ACK_IN=4, TEMP_F=5 }
			expected=0;

		if(a[0] == '@')
		{
			a++;
			silent=1;
		}

		// Delays and does nothing else.
		// i.e. D100 for 100 milliseconds, U10000 for 10000 microseconds
		if(sscanf(a,"%[dDuU]%d%c", &val, &pin, &toss) == 2)
		{
			if((val == 'd')||(val == 'D')) pin *= 1000L;

			INFO_PRINT("Delaying %d us\n", pin);
			if(pin>0)	usleep(pin);
			continue;
		}
		// Sends an ACK to the device and waits for it to reply.
		else if(strcmp(a, "ACK") == 0)
		{
			INFO_PRINT("Ping device\n");
			cmd=CMD_PING;
			expected=ACK_IN;
		}
		// Just a number, i.e. 3
		// Reads the logic value on the given pin
		else if(sscanf(a, "%d%c", &pin, &toss) == 1)
		{
			INFO_PRINT("Read value pin %d\n", pin);

			pincmd=digital_in;
			expected=PIN_IN;
		}

		//	V0	Read voltage on pin 0
		//	T0	Reads temp in Celsius on pin 0
		//	F0	Reads temp in Fahrenheit on pin 0
		//	Z0	Tristate pin 0
		else if(sscanf(a, "%[vVtTzZfF]%d%c", &val, &pin, &toss) == 2)
		{
			switch(val)
			{
			case 'v':	case 'V':
				INFO_PRINT("Read voltage pin %d\n", pin);

				pincmd=analog_in;
				expected=VOLT_IN;
				break;

			case 't':	case 'T':
			case 'f':	case 'F':
				INFO_PRINT("Read temp pin %d\n", pin);

				pincmd=temp_in;
				if((val == 'f') || (val == 'F'))
					expected=TEMP_F;
				else	expected=TEMP_C;
				break;

			case 'z':	case 'Z':
				INFO_PRINT("Tristate pin %d\n", pin);
				pincmd=digital_in;
				silent=1;
				expected=PIN_IN;
				break;
			default:
				break;
			}
		}
		// 0:1		Set pin 0's logic output to 1
		// 1:0		Set pin 1's logic output to 0
		else if(sscanf(a, "%d:%c%c", &pin, &val, &toss) == 2)
		{
			switch(val)
			{
			case '0':	// Set pin to gnd potential
				INFO_PRINT("Ground pin %d\n", pin);

				pincmd=set_lo;
				break;
			case '1':	// set pin to +5v
				INFO_PRINT("+5v pin %d\n", pin);

				pincmd=set_hi;
				break;
			default:
				fprintf(stderr, "Unknown command %s\n", a);
				continue;
				break;
			}
		}

		if(pincmd)
		{
			if((pin > 7 ) || (pin < 0))
			{
				fprintf(stderr, "Invalid pin %d\n", pin);
				continue;
			}

			cmd=pincmd[pin];
		}

RETRY_COMMAND:
		if(cmd)
		{
			if(write(fd, &cmd, 1) != 1)
				perror("Couldn't write command");
		}

		switch(expected)
		{
		unsigned short vdata;
		temp_t *tdata;
		unsigned char bdata;

		case TEMP_C:
		case TEMP_F:
			tdata=temp+temp_pos;
			read_n(fd, tdata, sizeof(temp_t));


			// Couldn't read a temperature at all
			if((unsigned short)(*tdata) == TEMP_ERROR)
			{
				// Don't bother trying more than twice
				// unless we get a POR value on second read
				if(tries > globals.trymax)
				{
					INFO_PRINT("No sensor detected\n");
					printf("ERR\n");
					continue;
				}

				DEBUG_PRINT(("No value.  First read?\n"));
				tries++;	// Time to retry.
				goto RETRY_COMMAND;
			}
			// Got a suspicious 0x0550, which is what
			// the sensor defaults to on power-on
			else if((unsigned short)(*tdata) == TEMP_RESET)
			{
				// If we get it on first try, assume
				// it's valid
				if(tries && (tries > globals.trymax))
				{
					INFO_PRINT("Timed out\n");
					printf("WAIT\n");
					continue;
				}
				else if(tries)
				{
					DEBUG_PRINT(("Got POR, Retry %d\n", tries));
					usleep(globals.tryms*1000L);
					tries++;
					goto RETRY_COMMAND;
				}
			}

			// If we make it this far, we think the temp's valid.
			if(!silent)
			{
				double v=BIN16_TO_TEMP(*tdata);

				// Out-of-range temps
				if(((*tdata) > TEMP_MAX) || ((*tdata) < TEMP_MIN))
					goto RETRY_COMMAND;

				// Have we read enough temperatures?
				// If not, reset tries, increase pos,
				// retry.
				if(temp_pos < (globals.rdmin-1))
				{
					tries=0;
					temp_pos++;
					goto RETRY_COMMAND;
				}

				// Sort temperatures, to find the median
				if(temp_pos > 0)
				qsort(temp, temp_pos+1, sizeof(temp_t),
					(compare_func)compare_s);

				// Point to the median value
				tdata=temp+(temp_pos/2);

				if(globals.raw)
					fwrite(tdata, 1, 2, stdout);
				else if(expected == TEMP_C)
					printf("%.2f\n", BIN16_TO_TEMP(*tdata));
				else	printf("%.2f\n", BIN16_TO_F(*tdata));
			}
			break;

		case VOLT_IN:
			read_n(fd, &vdata, sizeof(vdata));

			if(!silent)
			{
				if(globals.raw)
					fwrite(&vdata, 1, 2, stdout);
				else	printf("%.2fV\n", BIN16_TO_VOLT(vdata));
			}
			break;

		case PIN_IN:
			read_n(fd, &bdata, sizeof(bdata));

			if(!silent)
			{
				if(globals.raw)
					fwrite(&bdata, 1, 1, stdout);
				else	printf("%d\n", bdata);
			}
			break;

		case ACK_IN:
			read_n(fd, &bdata, sizeof(bdata));
			if(bdata == REPLY_PING)
				INFO_PRINT("Got ACK\n");

			if(bdata != REPLY_PING)
			{
				fprintf(stderr, "Ping failed, quitting\n");
				close(fd);
				return(3);
			}
		default:
			break;
		}
	}

#ifdef USE_LOCK
	if(globals.lock)
	{
		struct flock l;

		memset(&l, 0, sizeof(l));
		l.l_type=F_UNLCK;
		l.l_whence=SEEK_SET;
		l.l_start=0L;
		l.l_len=512L;

		if(fcntl(fd, F_SETLK, &l) != 0)
		{
			perror("Couldn't clear lock");
			close_dev(fd);
			return(1);
		}
	}
#endif

	close_dev(fd);
	return(0);
}

int close_dev(int fd)
{
	if(fd >= 0)
	{
		if(strcmp(globals.mode, "-") != 0)	// Change no TTY settings
		if(termios_restore(fd) < 0)
		fprintf(stderr, "Couldn't restore termios\n");

		close(fd);
	}
}

int open_dev(void)
{
	int fd=-1;

	fd=open(globals.dev, O_RDWR);

	if(fd < 0)
		return(-1);

	if(strcmp(globals.mode, "-") == 0)	// Change no TTY settings
		return(fd);

	if(termios_save(fd) < 0)
	{
		close(fd);
		return(-1);
	}

	if(termios_set(fd, globals.mode) < 0)
	{
		close(fd);
		return(-1);
	}

	return(fd);
}


static int compare_s(const temp_t *a, const temp_t *b)
{
	return((*a)-(*b));
}

const struct opt mopts[]={
	{"device",	'd',	1, &(globals.dev), NULL, "-d /dev/ttyUSB0" },
	{"mode",	'm',	1, &(globals.mode),NULL, "-m 115200N81 (- means leave alone)" },
	{"verbose",	'v',	0, &(globals.verbose), NULL, "Enables debug output" },
	{"help",	'h',	0, &(globals.usage), NULL, "Shows this listing" },
	{"raw",		'r',	0, &(globals.raw), NULL, "Outputs raw bytes"},
	{"tempms",	'm',	1, &(globals.tryms), "%d", 
			"Milliseconds between temp read retries" },
	{"retries",	'n',	1, &(globals.trymax), "%d", "Max # of tries to read a temp from a pin"},
	{"count",	'c',	1, &(globals.rdmin), "%d", "Min # of valid temperature reads per sensor, 1-7"},
	{"lock",	'l',	0, &(globals.lock), NULL, "Enables cooperative locking" },
	{NULL,		0,	0,	NULL,	 NULL, 	NULL }
	};
